from tinydb import TinyDB, Query, where
import pika
import json
import time
import requests
from .Network import NetworkHelpers
from flask import Flask, request
import threading


class Beacon:
  def __init__(self, project, config):
    print("Beacon Init")
    self.project = project
    self.config = config
    self.db = TinyDB(self.project.id+'.json')
    self.queues = {}
    self.devices = {}
    self.helper = NetworkHelpers()
    self.connection = None
    self.rabbitAutoconnect = False
    print("[Beacon] Setting up the routes...")
    self.setupServerRoutes()
    self.initRabbit()
  
  # Init the RabbitMQ connection
  def initRabbit(self):
    print("[Beacon] Scanning for RabbitMQ...")
    self.queueIP = self.helper.get_rabbitmq_local_ip()
    if self.queueIP is not None:
      print("[Beacon] RabbitMQ found on", self.queueIP)
      self.rabbitAutoconnect = False
      self.connectToRabbit()
    else:
      print("[Beacon] RabbitMQ not found")
      self.rabbitAutoconnect = True
    # Start the auto-connect thread
    self.serverThread = threading.Thread(target=self.rabbitAutoConnect, args=())
    self.serverThread.start()
  
  # Auto-connect to Rabbit if offline or disconnected (every 5 sec)
  def rabbitAutoConnect(self):
    while True:
      if self.rabbitAutoconnect:
        print("[Beacon] Scanning for RabbitMQ...")
        self.queueIP = self.helper.get_rabbitmq_local_ip()
        if self.queueIP is not None:
          self.rabbitAutoconnect = False
          self.connectToRabbit()
      time.sleep(5)

  # Connect to RabbitMQ
  def connectToRabbit(self):
    print("[Beacon] Connecting to RabbitMQ...")
    if self.queueIP is None:
      print("[Beacon] RabbitMQ not found")
      self.rabbitAutoconnect = True
    else:
      print("[Beacon] RabbitMQ found:", self.queueIP)
      try:
        self.connection = pika.BlockingConnection(pika.ConnectionParameters(self.queueIP))
        # Upload the buffered data if any
        self.emitBufferedData()
      except Exception as e:
        self.onRabbitClose()
  
  # In case Rabbit closes
  def onRabbitClose(self):
    if self.rabbitAutoconnect is not True:
      print("[Beacon] RabbitMQ closed")
      self.rabbitAutoconnect = True
      self.queues = {} # Clears the queues
      self.connection = None # Reset RabbitMQ
  
  # Start the Beacon
  def start(self):
    print("[Beacon] Looking for devices on the local network...")
    self.findServices()
    self.displayServices()
    print("[Beacon] Saying hello...")
    self.sayHello()
  
  # Say hello to announce our presence to other devices
  def sayHello(self):
    for deviceID in self.devices.keys():
      if deviceID != self.project.id:
        response = self.callService(self.devices[deviceID]["address"], self.devices[deviceID]["port"], "/mesh/hello", self.project.specs()[0])
        print("[Beacon] HELLO response: ", deviceID, response)
  
  # Execute a POST on one of the services
  def callService(self, address, port, endpoint, params=None):
    try:
      response = requests.post(f'http://{address}:{port}{endpoint}', data=params, timeout=5)
      print("[Raw response]", f'http://{address}:{port}{endpoint}', response)
      if response.status_code == 200:
        data = json.loads(response.text)
        return data
    except:
      return None

  # Find all the online devices on the network
  def findServices(self):
    self.devices = self.helper.find_emitters_and_consumers()
  
  # Display the services that are online
  def displayServices(self):
    for deviceID in self.devices.keys():
      print("[Beacon]["+self.devices[deviceID]["name"]+":"+str(self.devices[deviceID]["id"])+"] port", self.devices[deviceID]["port"])
    return self.devices
  
  # Add a service to our list
  def addService(self, serviceData):
    if serviceData["id"] not in self.devices.keys():
      self.devices[serviceData["id"]] = serviceData
      print("[Beacon] New device: ", serviceData["id"], serviceData["name"])
      self.displayServices()
    
  # Setup the mandatory Beacon routes
  def setupServerRoutes(self):
    self.project.server.createRoute("/specs", self.project.specs)
    self.project.server.createRoute("/ping", self.ping)
    self.project.server.createRoute("/mesh/hello", self.newMeshService)
  
  # On ping
  def ping(self):
    return self.project.specs()

  # On new device
  def newMeshService(self):
    service = request.form.to_dict()
    print("[Mesh Data]", service)
    self.addService(service)
    return {"ok": True}, 200
  
  # Get a queue by name
  def getQueue(self, name):
    if name not in self.queues.keys():
      self.queues[name] = self.connection.channel()
      self.queues[name].queue_declare(queue=name, durable=True)
    return self.queues[name]
  
  # Buffer data in a local db in case the queue is offline
  def bufferData(self, name, data):
    self.db.insert({'queue': name, 'data': data, "sent": False})
  
  # When RabbitMQ is back online, Emit buffered data
  def emitBufferedData(self):
    # Read the buffered data from TinyDB
    buffered_data = self.db.search(where('sent') == False)
    if len(buffered_data)>0:
      print("[Beacon] Starting upload of "+str(len(buffered_data))+" buffered items")
      # For each item in the buffered data
      doc_ids = []
      for item in buffered_data:
        queue_name = item['queue']
        data = item['data']
        # Emit the data
        self.emit(queue_name, data)
        doc_ids.append(item.doc_id)

        # Mark the data as sent
        #item['sent'] = True
        #self.db.update(item, doc_ids=[item.doc_id])

      self.db.remove(doc_ids=doc_ids)
      print("[Beacon] Buffer upload completed")
  
  # Emit data (public method)
  def emit(self, name, data, metas=None):
    if metas is None:
      metas = {}
    #print(name, data)
    data["serviceId"] = self.project.id
    self.pushQueue(name, data)

  # Push data to a queue (buffer if the queue is offline)
  def pushQueue(self, name, data):
    if self.connection is None:
      # No queue, we buffer
      print("[Beacon:buffer]", name, data)
      self.bufferData(name, data)
    try:
      channel = self.getQueue(name)
      channel.basic_publish(exchange='', routing_key=name, body=json.dumps(data))
      print("[Beacon:sent]", name, data)
    except Exception as e:
      self.onRabbitClose()
      print("[Beacon:buffer]", name, data)
      self.bufferData(name, data)




__version__ = '0.0.3'

print("Datahub Version :: ", __version__)

from .Project import Project
from .Beacon import Beacon
from .Beacon.Network import NetworkHelpers as Network
from .CTRL import Server

def hello_world():
    print("This is my first pip package!")



import requests
import socket
import json
from concurrent.futures import ThreadPoolExecutor
from tqdm import tqdm


class NetworkHelpers:
  def __init__(self):
    print("Network Helper")

  def get_local_ip_address(self):
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('8.8.8.8', 1))  # connect to a public IP address, any will do
    ip_address = s.getsockname()[0]
    s.close()
    return ip_address

  def find_emitters_and_consumers(self):
    """Finds all Emitters and Consumers on the local network."""
    ip_address = self.get_local_ip_address()
    ip_prefix = '.'.join(ip_address.split('.')[:-1]) + '.'
    emitters_and_consumers = {}
    with ThreadPoolExecutor() as executor:
      futures = [executor.submit(self.find_on_ip, ip_prefix + str(i)) for i in range(256)]
      with tqdm(total=len(futures), desc='Scanning network', unit=' IP address') as pbar:
        for future in futures:
          devices = future.result()
          if devices is not None:
            for device in devices:
              emitters_and_consumers[device["id"]] = device
            pbar.update(1)
    return emitters_and_consumers

  def find_on_ip(self, address):
    """Gets the RabbitMQ queues and description of an Emitter or Consumer."""
    result = []
    for port in range(8000, 8011):
      try:
        response = requests.get(f'http://{address}:{port}/specs', timeout=0.1)
        if response.status_code == 200:
          data = json.loads(response.text)
          if 'queues' in data and 'description' in data:
            data['address'] = address
            data['port'] = port
            result.append(data)
      except:
        pass
    if len(result)==0:
      return None
    return result

  def get_available_port(self):
    """Finds an available port to use for the local server."""
    port = 8000
    while True:
      with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        if s.connect_ex(('localhost', port)) != 0:
          return port
      port += 1

  def get_rabbitmq_local_ip(self):
    """Gets the local IP address of the machine running RabbitMQ."""
    local_ip = socket.gethostbyname(socket.gethostname())
    local_network_prefix = '.'.join(local_ip.split('.')[:-1]) + '.'

    open_ports = []
    #for i in tqdm(range(1, 255)):
    for i in range(1, 255):
      ip = local_network_prefix + str(i)
      for port in [5672]:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(0.1)
        result = sock.connect_ex((ip, port))
        if result == 0:
          open_ports.append({'ip': ip, 'port': port})
        sock.close()

    for p in open_ports:
      if p['port'] == 5672:
        return p['ip']

    return None




from flask import Flask, request
from flask_restful import Api
from werkzeug.serving import make_server
import threading

class Server:
  def __init__(self, project, config):
    self.config = config
    self.project = project
    self.app = Flask(__name__)
    self.api = Api(self.app)
    self.routes = {}
  
  def createRoute(self, path, callback):
    self.routes[path] = callback
  
  def handle_request(self, path=None):
    if path is None:
      path = request.path
    else:
      path = "/"+path
    if path not in self.routes:
      return {"error": True}, 404
    return self.routes[path]()
  
  def start(self, port):
    self.app.add_url_rule('/', 'handle_request', self.handle_request, methods=['GET', 'POST'])
    self.app.add_url_rule('/<path:path>', 'handle_request', self.handle_request, methods=['GET', 'POST'])
    self.server = make_server('0.0.0.0', port, self.app)

    self.serverThread = threading.Thread(target=self.server.serve_forever, args=())
    self.serverThread.start()
    print("[Server] Started on port", port)
    
  def stop(self):
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
      raise RuntimeError('Not running with the Werkzeug Server')
    func()


"""
def index():
  return {"hello": "world"}, 200

server = Server({}, {})
server.createRoute("/", index)
server.start(8000)
"""


from setuptools import setup, find_packages

from datahub_py import __version__

setup(
    name='datahub_py',
    version=__version__,
    description='Datahub core',

    url='https://github.com/26medias/Datahub-py',
    author='Julien L',
    author_email='julien@leap-forward.ca',

    packages=find_packages(exclude=['tests', 'tests.*']),

    extras_require={
        'flask': ['Flask==2.2.2', 'Flask_RESTful==0.3.9'],
        'numpy': ['numpy==1.23.5'],
        'requests': ['requests==2.28.1']
    },

    classifiers=[
        'Intended Audience :: Developers',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
)



from ..Beacon import Beacon
from ..CTRL import Server

import random
import string

class Project:
  def __init__(self, config):
    print("Project Init")
    self.config = config
    self.id = self.generateID()
    print("[Project] ID: "+self.id)
    self.server = Server(self, config)
    self.beacon = Beacon(self, config)
    self.port = self.beacon.helper.get_available_port()
    self.server.start(self.port)
    self.beacon.start()
  
  def generateID(self):
    return self.config.projectSpecs["type"]+(''.join(random.choices(string.ascii_uppercase + string.digits, k=7)))

  # Mandatory endpoint
  def specs(self):
    return {
      "id": self.id,
      "name": self.config.projectSpecs["name"],
      "queues": self.config.projectSpecs["queues"],
      "description": self.config.projectSpecs["description"],
      "address": self.beacon.helper.get_local_ip_address(),
      "port": self.port
    }, 200
